from os import getcwd

import typer
from git.exc import InvalidGitRepositoryError
from git.repo import Repo
from rich import print

from changelogger.app.manage import app as manage_app
from changelogger.app.unreleased import app as unrealeased_app
from changelogger.conf import settings

app = typer.Typer()
app.add_typer(manage_app)
app.add_typer(unrealeased_app)


@app.callback()
def changelogger():
    """Changelogger app help goes here."""
    if not settings.CHANGELOG_PATH.exists():
        print(
            "[bold red]Error: [/bold red]"
            f"Could not find changelog file [bold]{settings.CHANGELOG_PATH}[/bold]."
        )
        exit(1)

    if settings.HAS_DEFAULTS:
        try:
            _ = Repo(getcwd()).git_dir
        except InvalidGitRepositoryError:
            print(
                "[bold red]Error: [/bold red]"
                "Must be in a git repo to use the default behavior."
            )
            exit(1)



from rich import print
from rich.markdown import Markdown

from changelogger import changelog
from changelogger.exceptions import CommandException


def content(
    version: str,
    pretty: bool = True,
) -> None:
    """Retrieves the changelog content for the specified version."""
    all_versions = changelog.get_all_versions()
    if version not in all_versions:
        raise CommandException(f"Could not find version {version}.")

    i = all_versions.index(version)
    prev_version = (
        all_versions[i + 1] if i + 1 < len(all_versions) else "LINKS"
    )
    release_notes = changelog.get_release_notes(version, prev_version)

    md = release_notes.markdown()
    if pretty:
        print(Markdown(md))
    else:
        print(md)



from rich import print

from changelogger import changelog
from changelogger.exceptions import ValidationException


def check(
    sys_exit: bool = False,
) -> None:
    try:
        _check()
    except ValidationException as e:
        print(f"[bold red]Error:[/bold red] {str(e)}")
        if sys_exit:
            exit(1)
    else:
        print(
            ":white_heavy_check_mark: [bold green]All versioned files are valid![/bold green]"
        )


def _check() -> None:
    """Validates the specified versioned files are parsable and updatable."""
    # Validate there's at least 1 version
    # Point of Failure 0
    all_versions = changelog.get_all_versions()
    if not all_versions:
        raise ValidationException(
            "Expected there to be at least 1 version; None found."
        )

    # Validate all release notes are parseable for all versions
    # Point of Failure 1
    changelog_versions = ["Unreleased", *all_versions, "LINKS"]
    for version, prev_version in zip(
        changelog_versions, changelog_versions[1:]
    ):
        try:
            changelog.get_release_notes(version, prev_version)
        except:
            raise ValidationException(
                f"Failed to validate notes for version {version}"
            )

    # Validate there are links in the expected format for all versions
    # Point of Failure 2
    sorted_versions = changelog.get_sorted_versions()
    all_links = changelog.get_all_links()
    for prev_version, version in zip(sorted_versions, sorted_versions[1:]):
        link = all_links.get(version)
        if not link:
            raise ValidationException(
                f"Could not find the link for version {version}"
            )

        # Point of Failure 3
        if f"{prev_version}...{version}" not in link:
            raise ValidationException(
                f"Link is incorrect for version {version}"
            )

    link = all_links.get("Unreleased")
    # Point of Failure 4
    if not link:
        raise ValidationException(
            "Could not find the link for unreleased changes."
        )

    # Point of Failure 5
    if f"{all_versions[0]}...HEAD" not in link:
        raise ValidationException(
            "Link is incorrect for the unreleased changes."
        )

    # Point of Failure 6
    if sorted_versions[0] not in all_links:
        raise ValidationException(
            f"Could not find the link for version {sorted_versions[0]}"
        )



from importlib import resources
from pathlib import Path

DEFAULT_CHANGELOG_PATH = Path("CHANGELOG.md")

DEFAULT_OVERVIEW_JINJA_PATTERN = (
    r"### \[Unreleased\]([\s\S]*)### \[{{ old_version }}]"
)
DEFAULT_LINKS_JINJA_PATTERN = r"\[Unreleased\]:.*\n"

with (
    resources.path("assets", ".cl.overview.jinja2") as overview_path,
    resources.path("assets", ".cl.links.jinja2") as links_path,
):
    DEFAULT_OVERVIEW_JINJA_PATH = overview_path
    DEFAULT_LINKS_JINJA_PATH = links_path


CHANGELOGGER_PATH = Path(".changelogger.yml")



from collections.abc import Callable
from datetime import date
from typing import Any

from jinja2 import BaseLoader, Environment, Template

from changelogger.conf.models import VersionedFile
from changelogger.models.domain_models import ChangelogUpdate
from changelogger.utils import cached_compile


def _tmpl(jinja: str) -> Template:
    template_env = Environment(loader=BaseLoader())
    return template_env.from_string(jinja)


def _render_variables(
    versioned_file: VersionedFile,
    update: ChangelogUpdate,
) -> dict[str, Any]:
    return dict(
        new_version=update.new_version,
        old_version=update.old_version,
        today=date.today(),
        sections=update.release_notes.dict(),
        context=versioned_file.context,
    )


def update_with_jinja(
    file: VersionedFile,
) -> Callable:
    assert file.jinja or file.jinja_rel_path, "No valid jinja template found."

    replacement_str = file.jinja
    if not replacement_str and file.jinja_rel_path:
        replacement_str = file.jinja_rel_path.read_text()

    def update_fn(content: str, update: ChangelogUpdate) -> str:
        assert replacement_str
        render_kwargs = _render_variables(file, update)

        pattern = _tmpl(file.pattern).render(**render_kwargs)
        replacement = _tmpl(replacement_str).render(**render_kwargs)

        return cached_compile(pattern).sub(replacement, content)

    return update_fn
